package no.roek.nlpgraphs.detailedanalysis;

import com.mongodb.BasicDBList;
import com.mongodb.BasicDBObject;
import de.tudarmstadt.ukp.dkpro.lexsemresource.exception.LexicalSemanticResourceException;
import de.tudarmstadt.ukp.dkpro.lexsemresource.exception.ResourceLoaderException;
import de.tudarmstadt.ukp.similarity.algorithms.api.SimilarityException;
import no.roek.nlpgraphs.application.App;
import no.roek.nlpgraphs.document.PlagiarismPassage;
import no.roek.nlpgraphs.document.NLPSentence;
import no.roek.nlpgraphs.application.PlagiarismSearch;
import no.roek.nlpgraphs.graph.Graph;
import no.roek.nlpgraphs.misc.*;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedList;


public class PlagiarismFinder {

    private final double adjPlagTreshold;
    private final int cutoff;
    private Similarity ged;
    private double plagiarismThreshold;

    private DatabaseService db;
    private Map<String, Double> posEditWeights, deprelEditWeights;

    private Map<String, PlagiarismReference> plagRefCache;
    private Map<Integer, PlagiarismReference> plagRefMap;
    private Map<Integer,PlagiarismPassage> plagiarismPassages;
    private Map<Integer,Double> passageScores;
    

    public static Map<Integer,PlagiarismReference> positionMap= new HashMap<>();
   

    private static int n;
    private static int m;

   private static int previousRow;
   private static int previousColumn;
    

    private long cacheHit;
    private long uniqueHit;

    private Similarity sd;

    private CosineSimilarity cosMeasure;
    private NGramSimilarity oneGramMeasure;
    private NGramSimilarity twoGramMeasure;
    private NGramSimilarity threeGramMeasure;
    private NGramSimilarity oneGramJacMeasure;
    private NGramSimilarity twoGramJacMeasure;
    private NGramSimilarity threeGramJacMeasure;
    private StringTilingSimilarity tilingOneMeasure;
    private StringTilingSimilarity tilingTwoMeasure;
    private StringTilingSimilarity tilingThreeMeasure;
    private CommonSubseqSimilarity commonSubseqMeasure;
    private PairDistSimilarity pairDistMeasure;

    public PlagiarismFinder(DatabaseService db)
            throws LexicalSemanticResourceException, ResourceLoaderException, IOException {

        this.db = db;
        ConfigService cs = App.getGlobalConfig();
        plagiarismThreshold = cs.getPlagiarismThreshold();
        adjPlagTreshold = plagiarismThreshold + cs.adjPlagiarismTreshold();

        cutoff = cs.getCutoff();

        if (cs.getScoreType() == ConfigService.ScoreType.ALL ||
                cs.getScoreType() == ConfigService.ScoreType.COS ||
                cs.getScoreType() == ConfigService.ScoreType.ALL_LOGISTIC ||
                cs.getScoreType() == ConfigService.ScoreType.FAST ||
                cs.getScoreType() == ConfigService.ScoreType.FAST_GED ||
                cs.getScoreType() == ConfigService.ScoreType.FAST_SD_GED) {
            cosMeasure = new CosineSimilarity(db);
            oneGramMeasure = new NGramSimilarity(db, 1, false);
            twoGramMeasure = new NGramSimilarity(db, 2, false);
            threeGramMeasure = new NGramSimilarity(db, 3, false);

            oneGramJacMeasure = new NGramSimilarity(db, 1, true);
            twoGramJacMeasure = new NGramSimilarity(db, 2, true);
            threeGramJacMeasure = new NGramSimilarity(db, 3, true);

            tilingOneMeasure = new StringTilingSimilarity(db, 1);
            tilingTwoMeasure = new StringTilingSimilarity(db, 2);
            tilingThreeMeasure = new StringTilingSimilarity(db, 3);

            commonSubseqMeasure = new CommonSubseqSimilarity(db);
            pairDistMeasure = new PairDistSimilarity(db);
        }

        if (cs.getScoreType() == ConfigService.ScoreType.ALL ||
                cs.getScoreType() == ConfigService.ScoreType.SD ||
                cs.getScoreType() == ConfigService.ScoreType.SD_GED ||
                cs.getScoreType() == ConfigService.ScoreType.FAST_SD_GED) {
            sd = new CachedSimilarity(new SemanticDistance(db));
        }

        if (cs.getScoreType() == ConfigService.ScoreType.ALL ||
                cs.getScoreType() == ConfigService.ScoreType.GED ||
                cs.getScoreType() == ConfigService.ScoreType.SD_GED ||
                cs.getScoreType() == ConfigService.ScoreType.FAST_SD_GED) {
            ged = new CachedSimilarity(new GEDSimilarity(db));
        }
    }
    

      public static void printArray(double[][] matrix) {
    	    	
	    for (int row = 0; row < matrix.length; row++) {
	    	System.out.println("ROW NUMBER : "+row);
	        for (int column = 0; column < matrix[row].length; column++) {
	            System.out.print(matrix[row][column] + " ");
	        }
	        System.out.println();
	    }
	}

     //lagt til. E.
    public PlagiarismReference getReferenceFromPassage(PlagiarismPassage pass, double similarity){
    	
    	//String passageName= pass.toString();
    	//System.out.println("Passage name is: "+ passageName);
    	String sourceFile=pass.getTrainFile();
    	String suspiciousFile=pass.getTestFile();
    	int sourceSentence=pass.getTrainSentence();
    	int suspiciousSentence= pass.getTestSentence();
    	
        Graph source = GraphUtils.getGraph(db.getSentence(sourceFile, sourceSentence));
        Graph suspicious = GraphUtils.getGraph(db.getSentence(suspiciousFile, suspiciousSentence));    	
    	PlagiarismReference ref= XMLUtils.getPlagiarismReference(source, suspicious,similarity,true);
    	
    	return ref;
    }

         

              //method that uses the getScoreMatrix method

           public Map<Integer,Integer> findPlagiarism(PlagiarismJob job){
        	
                Map<Integer, Integer> rowColumn= new HashMap<>();
               
              List<PlagiarismPassage> passagelist= job.getTextPairs();
        	 if(passagelist.isEmpty()){
        		 return rowColumn;
        	 }

        	

        	 double[][] matrix=null;
        	  try {

        	       matrix= getScoreMatrix(job);
                       
                         } catch (SimilarityException e) {
                e.printStackTrace();
                     
                 }
        	
          n= matrix.length;
          m= matrix[0].length;      
          	
          	int row=0;
          	int column=0;
          	
          	for(int i=0; i<n;i++){
          		row++;
          		List<PlagiarismReference> rowReferences=null;
          		for (int j=0;j<m;j++){
          			column++;
          			if(matrix[i][j]<0.3){
          				PlagiarismPassage pass= plagiarismPassages.get((i*m)+j); 
          				PlagiarismReference ref= getReferenceFromPassage(pass,matrix[i][j]);
                      		        rowColumn.put(i, j); //rad, kolonne         			
          				positionMap.put((i*m)+j, ref);
                                             				
          			}
                 }          		         		
          	} 
          	
          	return rowColumn;
         }      

    
  

        
 public List<PlagiarismReference> mergePassages(Map<Integer,Integer> map){
      		
      		List<PlagiarismReference> mergedReferences= new ArrayList<>();
                PlagiarismReference ref= null;
                                       
                       		
      		      		
      		int row=0;
      		int column=0;    		
           
      		
      		for(Map.Entry<Integer,Integer> entry: map.entrySet()){                    
      		
      			row= entry.getKey();                                         
      			column=entry.getValue();
                     
                      
                        ref= positionMap.get(row*m +column);
                        int length= ref.getLengthInt();
                        int lengths= ref.getSourceLengthInt(); 
                        int off= ref.getOffsetInt();
                        int offs= ref.getSourceOffsetInt();              
                                                            
                                            	
                                                                                     

                         if((column-previousColumn <0)||(mergedReferences.isEmpty()) ||( row-previousRow >=30 && column-previousColumn >=30)){
                                      mergedReferences.add(ref);
                                     int offset= ref.getOffsetInt();                                                                                                              
                                    
                                     }
		 
      			      			
      			else if(row-previousRow < 30 && column-previousColumn < 30 && column-previousColumn !=0){
      				                               
                          PlagiarismReference ref0= mergedReferences.get(mergedReferences.size()-1);
                                   
                                mergedReferences.remove(ref0);
                                                                   
      				PlagiarismReference ref1= mergePassage(ref0, ref);                                                      
                                                        
      				mergedReferences.add(ref1);
                                                                 
      			           }
      			      						
      		previousRow=row;
      		previousColumn=column;
      		
      			
      		}
                      
      		
           		return mergedReferences;
      		
      	}
      	

      	
      	public static PlagiarismReference mergePassage(PlagiarismReference ref, PlagiarismReference other) {
      		if(ref.getOffsetInt() == other.getOffsetInt()) {
      			int len = Math.max(ref.getLengthInt(), other.getLengthInt());
      			ref.setLength(len);
      		}
      		if(ref.getSourceOffsetInt() == other.getSourceOffsetInt()) {
      			int len = Math.max(ref.getSourceLengthInt(), other.getSourceLengthInt());
      			ref.setSourceLength(len);
      		}
      		
      		if(ref.getOffsetInt() < other.getOffsetInt()) {
      			ref.setLength(other.getOffsetInt() + other.getLengthInt() - ref.getOffsetInt());
      		}else {
      			ref.setLength(ref.getOffsetInt() + ref.getLengthInt() - other.getOffsetInt());
      			ref.setOffset(other.getOffsetInt());
      		}

      		if(ref.getSourceOffsetInt() < other.getSourceOffsetInt()) {
      			ref.setSourceLength(other.getSourceOffsetInt() + other.getSourceLengthInt() - ref.getSourceOffsetInt());
      		}else {
      			ref.setSourceLength(ref.getSourceOffsetInt() + ref.getSourceLengthInt() - other.getSourceOffsetInt());
      			ref.setSourceOffset(other.getSourceOffsetInt());
      		}
      		
      		return ref;
      	}
      	
       
  
 

     public static <K extends Comparable,V extends Comparable> Map<K,V> sortByKeys(Map<K,V> map){
        List<K> keys = new LinkedList<K>(map.keySet());
        Collections.sort(keys);
      
        //LinkedHashMap will keep the keys in the order they are inserted
        //which is currently sorted on natural ordering
        Map<K,V> sortedMap = new LinkedHashMap<K,V>();
        for(K key: keys){
            sortedMap.put(key, map.get(key));
        }
      
        return sortedMap;
    }	


 public double[][] getScoreMatrix(PlagiarismJob job)throws SimilarityException{

      String jobname= job.getSuspFilename();
      List<PlagiarismPassage> passageList= job.getTextPairs();
     
     
    	plagiarismPassages = new HashMap<>();
        PlagiarismReference ref=null;
       
    	
    	PlagiarismPassage spyPassage= passageList.get(4);
    	String testFile= spyPassage.getTestFile();
    	String trainFile=spyPassage.getTrainFile();
    	
    	int n= getNrSentences(testFile);
    	int m= getNrSentences(trainFile);
    	    	
    	double[][] scoreMatrix= new double[n][m];    	
    
    	
    	
    	
    	             for(PlagiarismPassage pass: passageList){
    		 
      
    		//ref= getReferenceFromPassage(pass,0.0);
            //    int size1= (ref.getEndInt()-ref.getOffsetInt())+(ref.getSourceEndInt()-ref.getSourceOffsetInt());

    		int suspiciousSentence= pass.getTestSentence();    		
    		int sourceSentence= pass.getTrainSentence();    		
    		String sourceFile= pass.getTrainFile();
    		String suspiciousFile= pass.getTestFile();

              String suspId = suspiciousFile + "-" + suspiciousSentence;
              String srcId = sourceFile + "-" + sourceSentence;
	    	
    		    		    		
    		 double score=0.0;      

                // int size=350;
                                                       
              
               switch (App.getGlobalConfig().getScoreType()) {
               
              case GED:                   
               score = ged.getSimilarity(suspId,srcId);
                     break;
                        
              case SD:                               
               score = sd.getSimilarity(suspId,srcId);                                            	
                         break;         
                 
               case SD_GED:                
                score=(sd.getSimilarity(suspId,srcId)+ged.getSimilarity(suspId,srcId))/2;
                                      break;
                   
                      }     
                 
              
                scoreMatrix[suspiciousSentence-1][sourceSentence-1]= score;        
      		  int index=(suspiciousSentence-1)*m+(sourceSentence-1);
     		  plagiarismPassages.put(index,pass);
             
    	
       }
      
    	
               return scoreMatrix;
    	
       }
 


	 public int getNrSentences(String file){
    	
    	int nrSentences=0;
    	List<NLPSentence> sentenceList= db.getAllSentences(file);
    	nrSentences=sentenceList.size();
    	return nrSentences;
    	
    }


    public static PlagiarismReference mergeAdjacentReferences(PlagiarismReference ref1, PlagiarismReference ref2){

        int ref1_length= ref1.getLengthInt();
        int ref1_Offset= ref1.getOffsetInt();
        int ref1_sourceOffset = ref1.getSourceOffsetInt();
        int ref2_length= ref2.getLengthInt();

        int ref1_sourceLength= ref1.getSourceLengthInt();
        int ref2_sourceLength= ref2.getSourceLengthInt();

        String suspiciousFileName= ref1.getFilename();
        String sourceFileName= ref1.getSourceReference();
        String name = "detected-plagiarism";


        PlagiarismReference mergedPlagiarismReference =
                new PlagiarismReference(suspiciousFileName, name,ref1_Offset, ref1_length+ref2_length,
                        sourceFileName,ref1_sourceOffset,ref1_sourceLength+ref2_sourceLength);

        return mergedPlagiarismReference;
    }

    public String getNewLength(String offsetString, String lengthString, String newOffsetString, int ascending) {
        int offset = Integer.parseInt(offsetString);
        int len = Integer.parseInt(lengthString);
        int newOffset = Integer.parseInt(newOffsetString);

        int newLen =  len + ((offset - newOffset) * ascending);

        return Integer.toString(newLen);
    }

    /**
     * Only returns the plagiarism references from candidate retrieval.
     * Use this for measuring the candidate retrieval phase.
     */
    public List<PlagiarismReference> listCandidateReferences(PlagiarismJob job) {
        List<PlagiarismReference> plagReferences = new ArrayList<>();
        for (PlagiarismPassage pair : job.getTextPairs()) {
            Graph suspicious = GraphUtils.getGraph(db.getSentence(pair.getTestFile(), pair.getTestSentence()));
            Graph source = GraphUtils.getGraph(db.getSentence(pair.getTrainFile(), pair.getTrainSentence()));

            plagReferences.add(XMLUtils.getPlagiarismReference(source, suspicious, false));
        }

        return plagReferences;
    }

    public double getGEDSimilarity(BasicDBObject srcSent, BasicDBObject suspSent) {
        Graph source = GraphUtils.getGraph(srcSent);
        Graph suspicious = GraphUtils.getGraph(suspSent);
        GraphEditDistance ged = new GraphEditDistance(suspicious, source, posEditWeights, deprelEditWeights);
        return ged.getNormalizedDistance();
    }

    public double getSDSimilarity(BasicDBObject srcSent, BasicDBObject suspSent) {
        ArrayList<String> source_sem = SentenceUtils.getSentence(srcSent);
        ArrayList<String> suspicious_sem = SentenceUtils.getSentence(suspSent);

        return SemanticDistance.getSemanticDistance(source_sem, suspicious_sem);
    }

    public double[] getFastSimilarityVector(BasicDBObject srcSent, BasicDBObject suspSent) {
        List<String> suspTokens = SentenceUtils.getSentence(suspSent);
        List<String> srcTokens = SentenceUtils.getSentence(srcSent);

        double cosDist = cosMeasure.getSimilarity(suspTokens, srcTokens);
        double oneGramSim = oneGramMeasure.getSimilarity(suspTokens, srcTokens);
        double twoGramSim = twoGramMeasure.getSimilarity(suspTokens, srcTokens);
        double threeGramSim = threeGramMeasure.getSimilarity(suspTokens, srcTokens);
        double oneGramJacSim = oneGramJacMeasure.getSimilarity(suspTokens, srcTokens);
        double twoGramJacSim = twoGramJacMeasure.getSimilarity(suspTokens, srcTokens);
        double threeGramJacSim = threeGramJacMeasure.getSimilarity(suspTokens, srcTokens);
        double tilingOneSim = tilingOneMeasure.getSimilarity(suspTokens, srcTokens);
        double tilingTwoSim = tilingTwoMeasure.getSimilarity(suspTokens, srcTokens);
        double tilingThreeSim = tilingThreeMeasure.getSimilarity(suspTokens, srcTokens);
        double commonSubseqSim = commonSubseqMeasure.getSimilarity(suspTokens, srcTokens);
        double pairdistSim = pairDistMeasure.getSimilarity(suspTokens, srcTokens);

        return new double[]{ cosDist, oneGramSim, twoGramSim, threeGramSim, oneGramJacSim, twoGramJacSim,
                threeGramJacSim, tilingOneSim, tilingTwoSim, tilingThreeSim, commonSubseqSim, pairdistSim
        };
    }

    public double getFastSimilarity(BasicDBObject srcSent, BasicDBObject suspSent) {
        double[] fastSimVec = getFastSimilarityVector(srcSent, suspSent);

        double sim = 0.0;

        for (double s: fastSimVec) {
            sim += s;
        }

        return sim / fastSimVec.length;
    }

    public double getAllLogisticProb(BasicDBObject srcSent, BasicDBObject suspSent) {
        double gedSim = getGEDSimilarity(srcSent, suspSent);
        double sdDim = getSDSimilarity(srcSent, suspSent);
        double[] fastSimVec = getFastSimilarityVector(srcSent, suspSent);

        double[] f = new double[fastSimVec.length + 2];
        f[0] = gedSim;
        f[1] = sdDim;

        for (int i = 2; i < f.length; i++) {
            f[i] = fastSimVec[i-2];
        }

        double[] w = App.getGlobalConfig().getLogisticWeights();

        if (w.length != f.length + 1) {
            App.getLogger().warning("Inconsistent feature/weight vectors.");
            throw new RuntimeException();
        }

        double dot = w[f.length];

        for (int i = 0; i < f.length; i++) {
            dot += w[i]*f[i];
        }

        return 1 - (1 / (1 + Math.exp(-dot)));
    }

    public void saveCaches() {
        if ((ged != null) && (ged.getClass() == CachedSimilarity.class)) {
            CachedSimilarity cachedSim = (CachedSimilarity) ged;
            try {
                cachedSim.saveIfInvalidated();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        if ((sd != null) && (sd.getClass() == CachedSimilarity.class)) {
            CachedSimilarity cachedSim = (CachedSimilarity) sd;
            try {
                cachedSim.saveIfInvalidated();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
}
